import datetime
import decimal
import typing

from .version import PY_VERSION

if PY_VERSION >= (3, 9):
    from typing import Annotated  # type: ignore
else:
    from typing_extensions import Annotated  # type: ignore # pragma: no cover

if PY_VERSION >= (3, 10):
    from types import UnionType  # type: ignore # pragma: no cover
else:
    UnionType = None  # type: ignore # pragma: no cover

T = typing.TypeVar("T")
JsonDict = typing.Dict[str, typing.Any]

# This represents how avro.type is represneted in json.
AvroTypeRepr = typing.Union[JsonDict, typing.List, str]


class FieldInfo: ...


class Fixed: ...


class DecimalFieldInfo(FieldInfo):
    def __init__(self, max_digits: int = -1, decimal_places: int = 0) -> None:
        self.max_digits = max_digits
        self.decimal_places = decimal_places

    def __repr__(self) -> str:
        return f"DecimalFieldInfo(max_digits={self.max_digits}, decimal_places={self.decimal_places})"


class FixedFieldInfo(FieldInfo):
    def __init__(
        self,
        size: int,
        aliases: typing.Optional[typing.List[str]] = None,
        namespace=typing.Optional[str],
    ) -> None:
        self.size = size
        self.aliases = aliases
        self.namespace = namespace

    def __repr__(self) -> str:
        return f"FixedFieldInfo(size={self.size}, aliases={self.aliases}, namespace={self.namespace})"


class Int32FieldInfo(FieldInfo): ...


class Float32FieldInfo(FieldInfo): ...


class TimeMicroFieldInfo(FieldInfo): ...


class DateTimeMicro2FieldInfo(FieldInfo): ...


def confixed(
    *,
    size,
    aliases: typing.Optional[typing.List[str]] = None,
    namespace: typing.Optional[str] = None,
) -> typing.Type[bytes]:
    return Annotated[Fixed, FixedFieldInfo(size=size, aliases=aliases, namespace=namespace)]  # type: ignore[return-value]


def condecimal(*, max_digits: int, decimal_places: int) -> typing.Type[decimal.Decimal]:
    return Annotated[
        decimal.Decimal,
        DecimalFieldInfo(max_digits=max_digits, decimal_places=decimal_places),
    ]  # type: ignore[return-value]


Int32 = Annotated[int, Int32FieldInfo()]
Float32 = Annotated[float, Float32FieldInfo()]
TimeMicro = Annotated[datetime.time, TimeMicroFieldInfo()]
DateTimeMicro = Annotated[datetime.datetime, DateTimeMicro2FieldInfo()]

CUSTOM_TYPES = (
    Int32,
    Float32,
    TimeMicro,
    DateTimeMicro,
    condecimal,
    confixed,
)



import abc
import dataclasses
import json
import logging
import typing
from collections import OrderedDict

import inflection
from typing_extensions import get_args

from dataclasses_avroschema import utils

logger = logging.getLogger(__name__)


@dataclasses.dataclass  # type: ignore
class Field:
    __slots__ = (
        "name",
        "type",
        "default",
        "parent",
    )

    name: str
    type: typing.Any  # store the python primitive type
    parent: typing.Any
    default: typing.Any
    default_factory: typing.Any = dataclasses.MISSING
    exclude_default: bool = False
    inner_name: typing.Optional[str] = None
    metadata: typing.Dict = dataclasses.field(default_factory=dict)
    model_metadata: utils.SchemaMetadata = dataclasses.field(default_factory=utils.SchemaMetadata)
    extra_default_types_allowed: typing.Tuple = dataclasses.field(default_factory=tuple)

    def __post_init__(self) -> None:
        self.exclude_default = self.metadata.pop("exclude_default", False)  # type: ignore
        self.inner_name = self.metadata.pop("inner_name", None)  # type: ignore

    @property
    def avro_type(self) -> typing.Union[str, typing.Dict]:  # type: ignore
        ...  #  pragma: no cover

    @staticmethod
    def _get_self_reference_type(a_type: typing.Any) -> str:
        if getattr(a_type, "__args__", None):
            internal_type = a_type.__args__[0]
            return internal_type.__forward_arg__
        return a_type.__name__

    @staticmethod
    def get_singular_name(name: str) -> str:
        return inflection.singularize(name)

    def get_metadata(self) -> typing.List[typing.Tuple[str, str]]:
        return [(name, value) for name, value in self.metadata.items()]

    def render(self) -> OrderedDict:
        """
        Render the fields base on the avro field

        At least will have name and type.

        returns:
            OrderedDict(
                ("name", "a name"),
                ("type", "a type"),
                ("default", "default value")
            )

            The default key is optional.

            If self.type is:
                * list, the OrderedDict will contains the key items inside type
                * tuple, he OrderedDict will contains the key symbols inside type
                * dict, he OrderedDict will contains the key values inside type
        """
        template = OrderedDict(self.get_metadata() + [("name", self.name), ("type", self.get_avro_type())])
        default = self.get_default_value()

        if default is not dataclasses.MISSING:
            if default is not None:
                self.validate_default(default)
                default = self.default_to_avro(default)

            template["default"] = default

        return template

    def get_default_value(self) -> typing.Any:
        if self.exclude_default:
            return dataclasses.MISSING

        is_default_factory_callable = callable(self.default_factory)

        if is_default_factory_callable:
            return self.default_factory()
        return self.default

    def validate_default(self, default: typing.Any) -> bool:
        a_type = (self.type, *self.extra_default_types_allowed)
        msg = f'Invalid default type {type(default)} for field "{self.name}". Default should be {self.type}'

        if utils.is_annotated(self.type):
            a_type, _ = get_args(self.type)

        assert isinstance(default, a_type), msg
        return True

    def default_to_avro(self, value: typing.Any) -> typing.Any:
        return value

    def to_json(self) -> str:
        return json.dumps(self.render(), indent=2)

    def to_dict(self) -> dict:
        return dict(self.render())

    @abc.abstractmethod
    def get_avro_type(self) -> typing.Any: ...  # pragma: no cover

    def fake(self) -> typing.Any:
        return None

    def exist_type(self) -> int:
        # filter by the same field types
        same_types = [field.type for field in self.parent._user_defined_types if field.type == self.type]

        # If length > 0, means that it is the first appearance
        # of this type, otherwise exist already.
        return len(same_types)



import typing


class InvalidMap(Exception):
    def __init__(self, field_name: str, key_type: typing.Any) -> None:
        self.field_name = field_name
        self.key_type = key_type

    def __repr__(self) -> str:
        class_name = self.__class__.__name__  # pragma: no cover
        return f"{class_name} {self.field_name}"  # pragma: no cover

    def __str__(self) -> str:
        return f"Invalid map on field {self.field_name}. Keys must be string not {self.key_type}"


class InvalidSymbol(Exception):
    def __init__(self, field_name: str, symbol: str) -> None:
        self.field_name = field_name
        self.symbol = symbol

    def __repr__(self) -> str:
        class_name = self.__class__.__name__  # pragma: no cover
        return f"{class_name} {self.field_name}"  # pragma: no cover

    def __str__(self) -> str:
        return f"Symbol {self.symbol} does not match the regular expression [A-Za-z_][A-Za-z0-9_]*"



import copy
import dataclasses
import json
import typing
from dataclasses import dataclass, field
from string import Template

import casefy

from dataclasses_avroschema import serialization
from dataclasses_avroschema.fields import field_utils
from dataclasses_avroschema.types import AvroTypeRepr, JsonDict

from . import (
    avro_to_python_utils,
    templates,
)


class FieldRepresentation(typing.NamedTuple):
    name: str
    string_representation: str
    has_default: bool


@dataclass
class BaseGenerator:
    field_identation: str = "\n    "
    imports: typing.Set[str] = field(default_factory=set)
    imports_dict: typing.Dict[str, str] = field(default_factory=dict)

    # extras is used fot extra code that is generated, for example Enums
    extras: typing.List[str] = field(default_factory=list)

    metadata_fields_mapper: typing.Dict[str, str] = field(
        default_factory=lambda: {
            # doc is not included because it is rendered as docstrings
            "namespace": "namespace",
            "aliases": "aliases",
            "default": "default",
        }
    )
    metadata_field_templates: typing.Dict[str, Template] = field(
        default_factory=lambda: {
            "namespace": templates.metaclass_field_template,
            "doc": templates.metaclass_field_template,
            "aliases": templates.metaclass_alias_field_template,
            "original_schema": templates.metaclass_schema_field_template,
            "default": templates.metaclass_field_template,
            "schema_name": templates.metaclass_field_template,
        }
    )
    # represent the decorator to add in the base class
    base_class_decorator: str = ""
    base_class: str = field(init=False)
    field_template: Template = field(init=False)
    avro_type_to_lang: typing.Dict[str, str] = field(init=False)
    logical_types_imports: typing.Dict[str, str] = field(init=False)

    # Boolean to indicate whether original_schema field containing the original schema string should be generated in
    # Meta class of all generated objects
    include_original_schema: bool = False

    def __post_init__(self) -> None:
        self.avro_type_to_lang = avro_to_python_utils.AVRO_TYPE_TO_PYTHON
        self.logical_types_imports = avro_to_python_utils.LOGICAL_TYPES_IMPORTS

    def add_class_imports(self) -> None: ...

    def _resolve_type_from_metadata(self, *, field: JsonDict) -> typing.Optional[str]: ...

    def render(self, schemas: typing.List[JsonDict]) -> str:
        result = []

        for schema in schemas:
            if "fields" in schema:
                result.append(self.render_class(schema=schema))
            else:
                # If is not a Record then it is an Enum
                self.render_field(field=schema, model_name="")

        classes = "\n".join(result)
        imports = self.render_imports()
        extras = self.render_extras()

        return templates.module_template.safe_substitute(
            classes=classes,
            imports=imports,
            extras=extras,
        ).lstrip("\n")

    def render_imports(self) -> str:
        """
        Render the imports needed for the python classes.
        """
        # sort the imports
        list_imports = list(self.imports)
        list_imports.sort()

        imports = "\n".join([imp for imp in list_imports])
        return templates.imports_template.safe_substitute(imports=imports)

    def render_extras(self) -> str:
        return "".join([extra for extra in self.extras])

    def render_metaclass(
        self,
        *,
        schema: JsonDict,
        field_order: typing.Optional[typing.List[str]] = None,
        decorator: str = "",
        add_schema_name: bool = False,
    ) -> typing.Optional[str]:
        """
        Render Class Meta that contains the schema matadata
        """
        metadata = [
            self.metadata_field_templates[meta_avro_field].safe_substitute(name=meta_field, value=value)
            for meta_avro_field, meta_field in self.metadata_fields_mapper.items()
            if (value := schema.get(meta_avro_field))
        ]

        # Parse the entire schema string into the Meta class field "original_schema" for each class.
        # If self.include_original_schema is set to True.
        if self.include_original_schema:
            metadata.append(self._add_schema_to_metaclass(self.metadata_field_templates["original_schema"], schema))

        if field_order is not None:
            metadata.append(
                templates.metaclass_alias_field_template.safe_substitute(
                    name="field_order",
                    value=field_order,
                )
            )

        if add_schema_name:
            metadata.append(
                self.metadata_field_templates["schema_name"].safe_substitute(name="schema_name", value=schema["name"])
            )

        properties = self.field_identation.join(metadata)

        if properties:
            # some formating to remove identation at the end of the Class Meta to make it more compatible with black
            return self.field_identation.join(
                [
                    line
                    for line in templates.metaclass_template.safe_substitute(
                        properties=properties, decorator=decorator
                    ).split("\n")
                ]
            ).rstrip(self.field_identation)
        return None

    def render_docstring(self, *, docstring: typing.Optional[str]) -> str:
        """
        Render the module with the classes generated from the schema
        """
        if not docstring:
            return ""

        indented = self.field_identation + self.field_identation.join(docstring.splitlines())

        return f'{self.field_identation}"""{indented}{self.field_identation}"""'

    def render_class(self, *, schema: JsonDict) -> str:
        """
        Render the class generated from the schema
        """
        name: str = casefy.pascalcase(schema["name"])
        record_fields: typing.List[JsonDict] = schema["fields"]
        self.add_class_imports()

        # Sort the fields according whether it has a default value
        fields_representation: typing.List[FieldRepresentation] = [
            self.render_field(field=field, model_name=name) for field in record_fields
        ]

        fields_representation_copy = copy.deepcopy(fields_representation)
        fields_representation.sort(key=lambda field: 1 if field.has_default else 0)

        field_order = None
        if fields_representation_copy != fields_representation:
            field_order = [field.name for field in fields_representation_copy]

        rendered_fields_string = self.field_identation.join(
            field.string_representation for field in fields_representation
        )
        docstring = self.render_docstring(docstring=schema.get("doc"))

        rendered_class = templates.class_template.safe_substitute(
            name=name,
            decorator=self.base_class_decorator,
            base_class=self.base_class,
            fields=rendered_fields_string,
            docstring=docstring,
        )

        add_schema_name = name != schema["name"]
        class_metadata = self.render_metaclass(schema=schema, field_order=field_order, add_schema_name=add_schema_name)
        if class_metadata is not None:
            rendered_class += class_metadata

        return rendered_class

    def render_dataclass_field(self, properties: str) -> str:
        return self.field_template.safe_substitute(properties=properties)

    def render_field(self, field: JsonDict, model_name: str) -> FieldRepresentation:
        """
        Render an avro field.

        1. If the field is a native one, we just render it.
        2. If the field is a record, array, map, fixed, enum
            we need to render the field again (recursive)
        3. If the field is a LogicalType, it may not have the
            the `name` property and the type is a `native` one
        """
        name = self.generate_field_name(field)

        type: AvroTypeRepr = field["type"]
        default = field.get("default", dataclasses.MISSING)
        field_metadata = self.get_field_metadata(field)

        # This flag tells whether the field is array, map, fixed
        is_complex_type = has_default = False

        if self.is_logical_type(field=field):
            is_complex_type = True
            # override the type so it can be use to get the default value in case that is needed
            type = field.get("logicalType") or field["type"]["logicalType"]
            language_type = self.parse_logical_type(field=field)
        elif isinstance(type, dict):
            field_representation = self.render_field(field=type, model_name=model_name)
            language_type = field_representation.string_representation
            has_default = field_representation.has_default
        elif isinstance(type, list):
            language_type = self.parse_union(field_types=type, model_name=model_name)
        elif type == field_utils.ARRAY:
            is_complex_type = True
            language_type = self.parse_array(field=field, model_name=model_name)
        elif type == field_utils.MAP:
            is_complex_type = True
            language_type = self.parse_map(field=field, model_name=model_name)
        elif type == field_utils.ENUM:
            is_complex_type = True
            language_type = self.parse_enum(field=field)
            # We must set the default Enums type level default to dataclasses.MISSING
            # as it is set in the Meta class.
            # Check https://github.com/marcosschroh/dataclasses-avroschema/issues/665
            default = dataclasses.MISSING
        elif type == field_utils.FIXED:
            is_complex_type = True
            language_type = self.parse_fixed(field=field)
        elif type == field_utils.RECORD:
            record = f"\n{self.render_class(schema=field)}"
            is_complex_type = True
            self.extras.append(record)
            language_type = casefy.pascalcase(field["name"])
        else:
            # Native field
            language_type = self.get_language_type(type=type, model_name=model_name)

            # check if the language type must be replaced with some extra class
            # specified in the field metadata only after the native type was resolved
            type_from_metadata = self._resolve_type_from_metadata(field=field)
            language_type = type_from_metadata or language_type

        if is_complex_type or not name:
            # If the field is a complext type or
            # name is an empty string (it means that the type is a native type
            # with the form {"type": "a_primitive_type"}, example {"type": "string"})
            result = language_type
        else:
            result = templates.field_type_template.safe_substitute(name=name, type=language_type)

        # optional field attribute
        default_generated = self.get_field_default(
            field_type=type, default=default, name=name, field_metadata=field_metadata
        )

        if default_generated not in (
            dataclasses.MISSING,
            "",
        ):
            if type != field_utils.DECIMAL:
                result += templates.field_default_template.safe_substitute(default=default_generated)
            if default is not dataclasses.MISSING:
                has_default = True

        return FieldRepresentation(name=name, string_representation=result, has_default=has_default)

    @staticmethod
    def generate_field_name(field: JsonDict) -> str:
        field_name = field.get("name", "")

        if field_name and not field_name.isidentifier():
            valid_identifier = casefy.snakecase(field_name)
            if valid_identifier != field_name:
                aliases = field.get("aliases", [])

                if valid_identifier not in aliases:
                    aliases.append(field_name)
                    field["aliases"] = aliases

            return valid_identifier
        return field_name

    @staticmethod
    def is_logical_type(*, field: JsonDict) -> bool:
        if field.get("logicalType"):
            return True

        field_type = field["type"]
        return isinstance(field_type, dict) and field_type.get("logicalType") is not None

    def parse_logical_type(self, *, field: JsonDict) -> str:
        field_name = field.get("name")
        default = field.get("default")
        field = field if field_name is None or field.get("logicalType") else field["type"]
        logical_type = field["logicalType"]

        type_from_metadata = self._resolve_type_from_metadata(field=field)

        if type_from_metadata is not None:
            type = type_from_metadata
        elif logical_type == field_utils.DECIMAL:
            # this is a special case for logical types
            type = self.parse_decimal(field=field, default=default)
        elif logical_type not in self.logical_types_imports:
            # Then it is a custom logicalType, so we default to the native type
            type = self.get_language_type(type=field["type"])
        else:
            # add the logical type import
            self.imports.add(self.logical_types_imports[logical_type])
            type = self.avro_type_to_lang[logical_type]

        if field_name is not None:
            field_repr = templates.field_type_template.safe_substitute(name=field_name, type=type)
            return field_repr

        return type

    def parse_decimal(self, *, field: JsonDict, default: typing.Optional[str] = None) -> str:
        precision = field["precision"]
        scale = field["scale"]

        self.imports.add("from dataclasses_avroschema import types")
        field_repr = templates.decimal_type_template.safe_substitute(precision=precision, scale=scale)

        if default is not None:
            self.imports.add("import decimal")
            default = templates.decimal_template.safe_substitute(
                value=serialization.string_to_decimal(value=default, schema=field)
            )
            field_repr += templates.field_default_template.safe_substitute(default=default)
        return field_repr

    def parse_union(self, *, field_types: typing.List, model_name: str) -> str:
        """
        Parse an Avro union

        An union field is an array like ["null", "str", "int", ...]

        If the first element is `null` it means that the property is optional.
        The first item is the actual type of the field a AVRO specifies.

        Attributes:
            field_types: List of avro types
            model_name: name of the model that contains the `union`
        """

        # XXX: Maybe more useful in general
        def render_type(typ: str) -> str:
            if isinstance(typ, dict):
                field_representation = self.render_field(field=typ, model_name=model_name)
                return field_representation.string_representation
            else:
                return self.get_language_type(type=typ, model_name=model_name)

        if field_utils.NULL in field_types and len(field_types) == 2:
            # It is an optional field, we should include in the imports typing
            # and use the optional Template
            self.imports.add("import typing")
            (field_type,) = [f for f in field_types if f != field_utils.NULL]
            language_types = render_type(field_type)
            return templates.optional_template.safe_substitute(type=language_types)
        elif len(field_types) >= 2:
            # a union with more than 2 types
            self.imports.add("import typing")
            language_types_repr = ", ".join(render_type(t) for t in field_types)
            return templates.union_template.safe_substitute(type=language_types_repr)
        else:
            return render_type(field_types[0])

    def parse_array(self, field: JsonDict, model_name: str) -> str:
        """
        Parse an Avro array

        The type is is specify in the `items` attribute

        Example:
            {"name": "pets", "type": {"type": "array", "items": "string", "name": "pet"}}
        """
        type = field["items"]
        language_type = self._get_complex_langauge_type(type=type, model_name=model_name)

        return templates.list_template.safe_substitute(type=language_type)

    def parse_map(self, field: JsonDict, model_name: str) -> str:
        """
        Parse an Avro map

        The type is is specify in the `values` attribute. All the keys must be `string`

        Example:
            {"name": "accounts_money", "type": {"type": "map", "values": "float", "name": "accounts_money"}},
        """
        type = field["values"]
        language_type = self._get_complex_langauge_type(type=type, model_name=model_name)

        return templates.dict_template.safe_substitute(type=language_type)

    def parse_fixed(self, field: JsonDict) -> str:
        self.imports.add("from dataclasses_avroschema import types")
        properties = f"size={field['size']}"

        namespace = field.get("namespace")
        aliases = field.get("aliases")

        if namespace is not None:
            properties += f', namespace="{namespace}"'

        if aliases is not None:
            properties += f", aliases={aliases}"

        return templates.fixed_template.safe_substitute(properties=properties)

    def parse_enum(self, field: JsonDict) -> str:
        """
        Parse an Avro Enum

        Avro Enums are asociated with enum.Enum types in python.
        We need a template for it in order to create the Enum
        """
        self.imports.add("import enum")
        enum_name: str = casefy.pascalcase(field["name"])

        symbols_map = {}
        symbols: typing.List[str] = field["symbols"]
        for symbol in symbols:
            key = casefy.uppercase(symbol)
            if key in symbols:
                key = symbol
            symbols_map[key] = symbol

        symbols_repr = self.field_identation.join(
            [
                templates.enum_symbol_template.safe_substitute(key=key, value=f'"{value}"')
                for key, value in symbols_map.items()
            ]
        )

        docstring = self.render_docstring(docstring=field.get("doc"))
        enum_class = templates.enum_template.safe_substitute(name=enum_name, symbols=symbols_repr, docstring=docstring)
        add_schema_name = enum_name != field["name"]
        metaclass = self.render_metaclass(
            schema=field,
            decorator=templates.METACLASS_DECORATOR,
            add_schema_name=add_schema_name,
        )

        if metaclass:
            enum_class += metaclass

        self.extras.append(enum_class)

        return enum_name

    def _get_complex_langauge_type(self, *, type: typing.Any, model_name: str) -> str:
        """
        Get the language type for complex types (array and maps)
        """
        self.imports.add("import typing")

        if isinstance(type, dict):
            field_representation = self.render_field(field=type, model_name=model_name)
            language_type = field_representation.string_representation
        elif isinstance(type, list):
            language_type = self.parse_union(field_types=type, model_name=model_name)
        else:
            language_type = self.get_language_type(type=type, model_name=model_name)

        return language_type

    def get_language_type(
        self,
        *,
        type: str,
        default: typing.Optional[str] = None,
        model_name: typing.Optional[str] = None,
    ) -> str:
        if type in (field_utils.INT, field_utils.FLOAT):
            self.imports.add("from dataclasses_avroschema import types")

        if type == model_name:
            # it means that it is a one-to-self-relationship
            return templates.type_template.safe_substitute(type=type)
        elif type not in self.avro_type_to_lang:
            # it means the type points to the an already specified type so it contains the type name
            # with optional namespaces, e.g. my_namespace.users.User
            # In this case we should return the last part of the string
            return casefy.pascalcase(type.split(".")[-1])
        elif default is not None:
            return str(self.avro_type_to_lang.get(type, default))
        else:
            return str(self.avro_type_to_lang.get(type, type))

    def get_field_metadata(self, field: JsonDict) -> JsonDict:
        keys_to_ignore = [
            "name",
            "type",
            "default",
            "pydantic-class",
        ]

        type = field["type"]

        if type == field_utils.ARRAY:
            keys_to_ignore.append("items")
        elif type == field_utils.MAP:
            keys_to_ignore.append("values")
        elif type == field_utils.FIXED:
            keys_to_ignore.extend(["namespace", "aliases", "size"])
        elif type == field_utils.ENUM:
            keys_to_ignore.extend(["symbols", "namespace", "aliases", "doc"])
        elif type == field_utils.RECORD:
            keys_to_ignore.extend(["fields", "doc", "aliases", "namespace"])
        elif self.is_logical_type(field=field):
            keys_to_ignore.extend(
                [
                    "logicalType",
                ]
            )

        return {name: value for name, value in field.items() if name not in keys_to_ignore}

    def get_field_default(
        self,
        *,
        field_type: AvroTypeRepr,
        default: typing.Any,
        name: str,
        field_metadata: typing.Optional[JsonDict] = None,
    ) -> typing.Any:
        """
        Returns the default value according to the field type

        TODO: docstrings

        Example:
            If the default is "bond" the method should return '"bond"' so the double quotes
            won't be scaped during the field render
        """
        field_metadata = field_metadata or {}
        field_metadata_repr = None
        default_repr = ""

        if isinstance(field_type, dict):
            inner_type = field_type["type"]
            inner_name = field_type.get("name", "")

            # Check that internal field name is different to the original name
            # for example: {"name": "age", "type": { "type": "array", "items": "string", "name": "my_age" }}
            # in this case my_age is not in age which means that this must be reflected in the model
            # This does not applu to `enums` and `records`
            if inner_name not in name and inner_type not in (
                field_utils.ENUM,
                field_utils.RECORD,
            ):
                field_metadata["inner_name"] = inner_name

            inner_default_repr = self.get_field_default(field_type=inner_type, name=inner_name or name, default=default)

            if inner_default_repr not in (None, dataclasses.MISSING):
                default_repr += inner_default_repr
        elif default is dataclasses.MISSING:
            pass
        elif field_type in (
            field_utils.STRING,
            field_utils.UUID,
        ):
            default_repr = f'"{default}"'
        elif field_type in (
            field_utils.BYTES,
            field_utils.FIXED,
        ):
            default_repr = f'b"{default}"'
        elif field_type == field_utils.ENUM:
            default_repr = f"{casefy.pascalcase(name)}.{casefy.uppercase(default)}"
        elif isinstance(field_type, list):
            # union type
            default_repr = self.get_field_default(field_type=field_type[0], default=default, name=name)
        elif isinstance(default, (dict, list)):
            # Then is can be a regular dict as default or a record
            if default:
                if field_type not in field_utils.AVRO_TYPES:
                    # Try to get the last part in case that the type is namespaced `types.bus_type.Bus`
                    field_type = field_type.split(".")[-1]
                    default = templates.instance_template.safe_substitute(type=field_type, properties=f"**{default}")

                return f"{default}"
            return default.__class__.__name__
        elif field_type in avro_to_python_utils.LOGICAL_TYPES_TO_PYTHON:
            func = avro_to_python_utils.LOGICAL_TYPES_TO_PYTHON[field_type]
            python_type = func(default)
            template_func = avro_to_python_utils.LOGICAL_TYPE_TEMPLATES[field_type]
            default_repr = template_func(python_type)
        else:
            default_repr = str(default)

        dataclass_field_properties = []
        if field_metadata:
            field_metadata_repr = f"metadata={field_metadata}"

        if field_metadata_repr or isinstance(default, (dict, list)):
            dataclass_field_properties = [field_metadata_repr]

            if isinstance(default, (dict, list)):
                if default:
                    dataclass_prop = f"default_factory=lambda: {default_repr}"
                else:
                    dataclass_prop = f"default_factory={default_repr}"

                dataclass_field_properties.append(dataclass_prop)
            else:
                if default is not dataclasses.MISSING:
                    dataclass_field_properties.append(f"default={default_repr}")

            default_repr = self.render_dataclass_field(
                properties=", ".join([prop for prop in dataclass_field_properties if prop])
            )

        return default_repr

    @staticmethod
    def _add_schema_to_metaclass(schema_template: Template, schema: JsonDict) -> str:
        """
        Parses provided schema to ModelGenerator.render() to a Meta field string by using
        metaclass_schema_field_template.
        """
        return schema_template.safe_substitute(name="original_schema", schema=json.dumps(schema))



import dataclasses
import datetime
import decimal
import enum
import inspect
import random
import re
import typing
import uuid

from typing_extensions import get_args, get_origin

from dataclasses_avroschema import (
    exceptions,
    schema_generator,
    serialization,
    types,
    utils,
    version,
)
from dataclasses_avroschema.faker import fake
from dataclasses_avroschema.utils import is_pydantic_model

from . import field_utils
from .base import Field

if version.PY_VERSION >= (3, 9):  # pragma: no cover
    GenericAlias = (
        typing.GenericAlias,  # type: ignore
        typing._GenericAlias,  # type: ignore
        typing._SpecialGenericAlias,  # type: ignore
        typing._UnionGenericAlias,  # type: ignore
    )  # noqa: E501
else:
    GenericAlias = typing._GenericAlias  # type: ignore  # pragma: no cover


__ALL__ = [
    "ImmutableField",
    "StringField",
    "IntField",
    "LongField",
    "BooleanField",
    "DoubleField",
    "FloatField",
    "BytesField",
    "NoneField",
    "ContainerField",
    "ListField",
    "TupleField",
    "DictField",
    "UnionField",
    "LiteralField",
    "FixedField",
    "EnumField",
    "SelfReferenceField",
    "DateField",
    "DatetimeField",
    "DatetimeMicroField",
    "TimeMilliField",
    "TimeMicroField",
    "UUIDField",
    "DecimalField",
    "RecordField",
    "AvroField",
]


class ImmutableField(Field):
    def get_avro_type(
        self,
    ) -> typing.Union[str, typing.List, typing.Dict[str, typing.Any]]:
        if self.default is None:
            return [field_utils.NULL, self.avro_type]
        return self.avro_type


@dataclasses.dataclass
class StringField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.STRING

    def fake(self) -> str:
        return fake.pystr()


@dataclasses.dataclass
class IntField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.INT

    def fake(self) -> int:
        return fake.pyint()


@dataclasses.dataclass
class LongField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.LONG

    def fake(self) -> int:
        return fake.pyint()


@dataclasses.dataclass
class BooleanField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.BOOLEAN

    def fake(self) -> bool:
        return fake.pybool()


@dataclasses.dataclass
class DoubleField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.DOUBLE

    def __post_init__(self):
        super().__post_init__()
        self.extra_default_types_allowed = (int,)

    def fake(self) -> float:
        return fake.pyfloat()


@dataclasses.dataclass
class FloatField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.FLOAT

    def fake(self) -> float:
        return fake.pyfloat()  # Roughly the range on a float32


@dataclasses.dataclass
class BytesField(ImmutableField):
    avro_type: typing.ClassVar[str] = field_utils.BYTES

    def default_to_avro(self, item: bytes) -> str:
        return item.decode()

    def fake(self) -> bytes:
        return fake.pystr().encode()


@dataclasses.dataclass
class NoneField(ImmutableField):
    @property
    def avro_type(self) -> str:
        return field_utils.NULL

    def get_avro_type(self) -> str:
        return self.avro_type


@dataclasses.dataclass
class ContainerField(Field):
    @property
    def avro_type(self) -> typing.Dict:  # type: ignore
        ...  # pragma: no cover

    def get_avro_type(self) -> types.JsonDict:
        avro_type = self.avro_type
        avro_type["name"] = self.inner_name or self.get_singular_name(self.name)

        return avro_type


@dataclasses.dataclass
class BaseListField(ContainerField):
    items_type: typing.Any = None
    internal_field: Field = dataclasses.field(init=False)

    @property
    def avro_type(self) -> typing.Dict:
        self.generate_items_type()
        return {"type": field_utils.ARRAY, "items": self.items_type}

    def get_default_value(self) -> typing.Union[typing.List, dataclasses._MISSING_TYPE]:
        default = super().get_default_value()

        if default is not dataclasses.MISSING:
            if default is None:
                return []
            else:
                if isinstance(default, tuple):
                    default = list(default)
        return default

    def validate_default(self, default):
        assert isinstance(default, list), f"List is required as default for field {self.name}"
        if not isinstance(self.internal_field, UnionField):
            for element in default:
                self.internal_field.validate_default(element)

    def default_to_avro(self, values: typing.List):
        return [self.internal_field.default_to_avro(item) for item in values]

    def generate_items_type(self) -> typing.Any:
        # because avro can have only one type, we take the first one
        items_type = self.type.__args__[0]

        if utils.is_union(items_type):
            self.internal_field = UnionField(
                self.name,
                items_type,
                default=self.default,
                default_factory=self.default_factory,
                model_metadata=self.model_metadata,
                parent=self.parent,
            )
        else:
            self.internal_field = AvroField(
                self.name,
                items_type,
                model_metadata=self.model_metadata,
                parent=self.parent,
            )

        self.items_type = self.internal_field.get_avro_type()


@dataclasses.dataclass
class ListField(BaseListField):
    def fake(self) -> typing.List:
        return [self.internal_field.fake()]


@dataclasses.dataclass
class TupleField(BaseListField):
    """
    This behaves on the same way as `ListField` with
    as in avro schema does not exist `tuples`

    The reason to have this is to generate a proper `fake`
    """

    def fake(self) -> typing.Tuple:
        return (self.internal_field.fake(),)


@dataclasses.dataclass
class DictField(ContainerField):
    values_type: typing.Any = None
    internal_field: typing.Any = None

    def __post_init__(self) -> None:
        super().__post_init__()
        key_type = self.type.__args__[0]

        if not issubclass(key_type, str):
            raise exceptions.InvalidMap(self.name, key_type)

    @property
    def avro_type(self) -> types.JsonDict:
        self.generate_values_type()
        return {"type": field_utils.MAP, "values": self.values_type}

    def get_default_value(
        self,
    ) -> typing.Union[types.JsonDict, dataclasses._MISSING_TYPE]:
        default = super().get_default_value()

        if default is not dataclasses.MISSING:
            if default is None:
                return {}
            else:
                assert isinstance(default, dict), f"Dict is required as default for field {self.name}"
        return default

    def validate_default(self, default):
        assert isinstance(default, dict)
        if not isinstance(self.internal_field, UnionField):
            for element in default.values():
                self.internal_field.validate_default(element)

    def default_to_avro(self, default: typing.Dict):
        return {key: self.internal_field.default_to_avro(value) for key, value in default.items()}

    def generate_values_type(self) -> typing.Any:
        """
        Process typing.Dict. Avro assumes that the key of a map is always a string,
        so we take the second argument to determine the value type
        """
        values_type = self.type.__args__[1]
        self.internal_field = AvroField(
            self.name,
            values_type,
            model_metadata=self.model_metadata,
            parent=self.parent,
        )
        self.values_type = self.internal_field.get_avro_type()

    def fake(self) -> typing.Dict[str, typing.Any]:
        # return a dict of one element with the items type specified
        return {fake.pystr(): self.internal_field.fake()}


@dataclasses.dataclass
class UnionField(Field):
    unions: typing.List = dataclasses.field(default_factory=list)
    internal_fields: typing.List[Field] = dataclasses.field(default_factory=list)
    elements: typing.Tuple = dataclasses.field(default_factory=tuple)

    def generate_unions_type(self) -> typing.List:
        """
        Generate union.

        Arguments:
            elements (typing.List): List of python types
            default (typing.Any): Default value
            default factory (typing.Calleable): Callable to get the default value for
                a list or dict type

        Returns:
            typing.List: List of avro types
        """
        self.elements = get_args(self.type)
        name = self.get_singular_name(self.name)

        unions: typing.List = []

        # Place default at front of list
        default_type = default_field = None
        if self.default is None and self.default_factory is dataclasses.MISSING:
            unions.insert(0, field_utils.NULL)
        elif type(self.default) is not dataclasses._MISSING_TYPE:
            default_type = type(self.default)
            default_field = AvroField(
                name,
                default_type,
                model_metadata=self.model_metadata,
                parent=self.parent,
            )
            unions.append(default_field.get_avro_type())
            self.internal_fields.append(default_field)

        for element in self.elements:
            # create the field and get the avro type
            field = AvroField(name, element, model_metadata=self.model_metadata, parent=self.parent)
            avro_type = field.get_avro_type()

            if avro_type not in unions and field != default_field:
                unions.append(avro_type)
                self.internal_fields.append(field)

        return unions

    def get_avro_type(self) -> typing.List:
        self.unions = self.generate_unions_type()
        return self.unions

    def default_to_avro(self, default: typing.Any):
        first_type = self.internal_fields[0]
        return first_type.default_to_avro(default)

    def validate_default(self, default):
        first_type = self.internal_fields[0]
        first_type.validate_default(default)

    def fake(self) -> typing.Any:
        # get a random internal field and return a fake value
        field = random.choice(self.internal_fields)
        return field.fake()


@dataclasses.dataclass
class LiteralField(Field):
    avro_field: Field = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        """
        Derives Avro schema type[s] and validation requirements
        """
        super().__post_init__()
        args = get_args(self.type)

        # always convert to an Union. If there is Literal has 1 argument then
        # the Union will handle it properly.
        native_type = typing.Union[tuple(type(a) for a in args)]  # type: ignore

        if native_type is str and self.model_metadata.convert_literal_to_enum:
            # if all args are string then we use a `StrEnum` as native type
            native_type = enum.Enum(self.name, {member: member for member in args})  # type: ignore

        self.avro_field = AvroField(
            name=self.name,
            native_type=native_type,
            parent=self.parent,
            default=self.default,
            default_factory=self.default_factory,
            model_metadata=self.model_metadata,
        )

    def get_avro_type(self) -> types.JsonDict:
        return self.avro_field.get_avro_type()

    def default_to_avro(self, default: typing.Any):
        return self.avro_field.default_to_avro(default)

    def validate_default(self, default: typing.Any) -> bool:
        return self.avro_field.validate_default(default)

    def fake(self) -> typing.Any:
        return random.choice(get_args(self.type))


@dataclasses.dataclass
class FixedField(BytesField):
    field_info: typing.Optional[types.FixedFieldInfo] = None
    size: int = 0
    aliases: typing.Optional[typing.List[str]] = None
    namespace: typing.Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        self.set_fixed()

    def set_fixed(self) -> None:
        if self.field_info is not None:
            self.size = self.field_info.size
            self.aliases = self.field_info.aliases
            self.namespace = self.field_info.namespace

    def get_avro_type(self) -> types.JsonDict:
        avro_type = {
            "type": field_utils.FIXED,
            "name": self.inner_name or self.get_singular_name(self.name),
            "size": int(self.size),
        }

        if self.namespace is not None:
            avro_type["namespace"] = self.namespace

        if self.aliases is not None:
            avro_type["aliases"] = self.aliases

        return avro_type

    def validate_default(self, default) -> bool:
        msg = "Invalid default type. Default should be bytes"
        assert isinstance(default, bytes), msg
        return True

    def fake(self) -> bytes:
        return fake.pystr(max_chars=self.size).encode()


@dataclasses.dataclass
class EnumField(Field):
    SYMBOL_REGEX: typing.ClassVar = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")
    symbols: typing.List[str] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.extra_default_types_allowed = (str,)
        self.symbols = self.get_symbols()
        self.validate_symbols()

    def _get_meta_class_attributes(self) -> typing.Dict[str, typing.Any]:
        # get Enum members
        members = self.type.__members__
        meta = members.get("Meta") or getattr(self.type, "Meta", None)

        doc: typing.Optional[str] = self.type.__doc__

        # On python < 3.11 Enums have a default documentation so we remove it
        if version.PY_VERSION < (3, 11) and doc == "An enumeration.":
            doc = None

        try:
            if meta is not None:
                meta = meta.value
        except AttributeError:
            pass

        metadata = utils.FieldMetadata.create(meta)
        if doc is not None:
            metadata.doc = doc.strip()

        return metadata.to_dict()

    def get_symbols(self) -> typing.List[str]:
        return [member.value for member in self.type if member.name != "Meta"]

    def validate_symbols(self) -> bool:
        for symbol in self.symbols:
            if not isinstance(symbol, str) or not self.SYMBOL_REGEX.fullmatch(symbol):
                raise exceptions.InvalidSymbol(field_name=self.name, symbol=symbol)

        return True

    def get_avro_type(self) -> typing.Union[str, types.JsonDict]:
        metadata = self._get_meta_class_attributes()
        # If the enum name in avro schema different than the python enum class name, use the schema_name.
        name = metadata.pop("schema_name", self.type.__name__)

        if not self.exist_type():
            user_defined_type = utils.UserDefinedType(name=name, type=self.type)
            self.parent._user_defined_types.add(user_defined_type)
            return {
                "type": field_utils.ENUM,
                "name": name,
                "symbols": self.symbols,
                **metadata,
            }
        else:
            namespace = metadata.get("namespace")
            if namespace is None:
                return name
            else:
                return f"{namespace}.{name}"

    def default_to_avro(self, default: typing.Any) -> typing.Any:
        if isinstance(default, str):
            return default
        return default.value

    def fake(self) -> typing.Any:
        return random.choice(self.get_symbols())


@dataclasses.dataclass
class SelfReferenceField(Field):
    def get_avro_type(self) -> typing.Union[typing.List[str], str]:
        str_type = self._get_self_reference_type(self.type)

        if self.default is None:
            # means that default value is None
            return [field_utils.NULL, str_type]
        return str_type

    def get_default_value(self) -> typing.Union[dataclasses._MISSING_TYPE, None]:
        # Only check for None because self reference default value can be only None
        if self.default is None:
            return None
        return dataclasses.MISSING

    def fake(self) -> typing.Any:
        if getattr(self.type, "__args__", None):
            # It means that self.type is `typing.Type['AType']`, and the argument is a string
            # then we return None
            return None
        return self.type.fake()


@dataclasses.dataclass
class DateField(ImmutableField):
    """
    The date logical type represents a date within the calendar,
    with no reference to a particular time zone or time of day.

    A date logical type annotates an Avro int, where the int stores
    the number of days from the unix epoch, 1 January 1970 (ISO calendar).
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_DATE

    def default_to_avro(self, date: datetime.date) -> int:
        """
        Convert to datetime and get the amount of days
        from the unix epoch, 1 January 1970 (ISO calendar)
        for a given date

        Arguments:
            date (datetime.date)

        Returns:
            int
        """
        date_time = datetime.datetime.combine(date, datetime.datetime.min.time())
        ts = (date_time - utils.epoch_naive).total_seconds()

        return int(ts / (3600 * 24))

    def fake(self) -> datetime.date:
        return fake.date_object()


@dataclasses.dataclass
class TimeMilliField(ImmutableField):
    """
    The time-millis logical type represents a time of day,
    with no reference to a particular calendar,
    time zone or date, with a precision of one millisecond.

    A time-millis logical type annotates an Avro int,
    where the int stores the number of milliseconds after midnight, 00:00:00.000.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_TIME_MILIS

    def default_to_avro(self, time: datetime.time) -> int:
        """
        Returns the number of milliseconds after midnight, 00:00:00.000
        for a given time object

        Arguments:
            time (datetime.time)

        Returns:
            int
        """
        hour, minutes, seconds, microseconds = (
            time.hour,
            time.minute,
            time.second,
            time.microsecond,
        )

        return int((((hour * 60 + minutes) * 60 + seconds) * 1000) + (microseconds / 1000))

    def fake(self) -> datetime.time:
        return fake.time_object()


@dataclasses.dataclass
class TimeMicroField(ImmutableField):
    """
    The time-micros logical type represents a time of day,
    with no reference to a particular calendar,
    time zone or date, with a precision of one millisecond.

    A time-micros logical type annotates an Avro long,
    where the int stores the number of milliseconds after midnight, 00:00:00.000000.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_TIME_MICROS

    def default_to_avro(self, time: datetime.time) -> float:
        """
        Returns the number of microseconds after midnight, 00:00:00.000000
        for a given time object

        Arguments:
            time (datetime.time)

        Returns:
            int
        """
        hour, minutes, seconds, microseconds = (
            time.hour,
            time.minute,
            time.second,
            time.microsecond,
        )

        return int((((hour * 60 + minutes) * 60 + seconds) * 1000000) + microseconds)

    def fake(self) -> datetime.time:
        datetime_object: datetime.datetime = fake.date_time(tzinfo=datetime.timezone.utc)
        datetime_object = datetime_object + datetime.timedelta(microseconds=random.randint(0, 999))
        return datetime_object.time()


@dataclasses.dataclass
class DatetimeField(ImmutableField):
    """
    The timestamp-millis logical type represents an instant on the global timeline,
    independent of a particular time zone or calendar, with a precision of one millisecond.

    A timestamp-millis logical type annotates an Avro long,
    where the long stores the number of milliseconds from the unix epoch,
    1 January 1970 00:00:00.000 UTC.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_DATETIME_MILIS

    def default_to_avro(self, date_time: datetime.datetime) -> int:
        """
        Returns the number of milliseconds from the unix epoch,
        1 January 1970 00:00:00.000 UTC for a given datetime
        """
        if date_time.tzinfo:
            ts = (date_time - utils.epoch).total_seconds()
        else:
            ts = (date_time - utils.epoch_naive).total_seconds()

        return int(ts * 1000)

    def fake(self) -> datetime.datetime:
        return fake.date_time(tzinfo=datetime.timezone.utc)


@dataclasses.dataclass
class DatetimeMicroField(ImmutableField):
    """
    The timestamp-millis logical type represents an instant on the global timeline,
    independent of a particular time zone or calendar, with a precision of one millisecond.

    A timestamp-millis logical type annotates an Avro long,
    where the long stores the number of milliseconds from the unix epoch,
    1 January 1970 00:00:00.000000 UTC.
    """

    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_DATETIME_MICROS

    def default_to_avro(self, date_time: datetime.datetime) -> int:
        """
        Returns the number of milliseconds from the unix epoch,
        1 January 1970 00:00:00.000000 UTC for a given datetime
        """
        if date_time.tzinfo:
            ts = (date_time - utils.epoch).total_seconds()
        else:
            ts = (date_time - utils.epoch_naive).total_seconds()

        return int(ts * 1000000)

    def fake(self) -> datetime.datetime:
        datetime_object: datetime.datetime = fake.date_time(tzinfo=datetime.timezone.utc)
        return datetime_object + datetime.timedelta(microseconds=random.randint(0, 999))


@dataclasses.dataclass
class UUIDField(ImmutableField):
    @property
    def avro_type(self) -> typing.Dict:
        return field_utils.LOGICAL_UUID

    def validate_default(self, default) -> bool:
        msg = f"Invalid default type. Default should be {str} or {uuid.UUID}"
        assert isinstance(default, (str, uuid.UUID)), msg

        return True

    def default_to_avro(self, uuid: uuid.UUID) -> str:
        return str(uuid)

    def fake(self) -> uuid.UUID:
        return uuid.uuid4()


@dataclasses.dataclass
class DecimalField(Field):
    field_info: typing.Optional[types.DecimalFieldInfo] = None
    max_digits: int = -1  # amount of digits, in avro is called `precision`
    decimal_places: int = 0  # amount of digits to the right side, in avro is called `scale`

    def __post_init__(self) -> None:
        super().__post_init__()
        self.set_precision_scale()

    def set_precision_scale(self) -> None:
        if self.field_info is not None:
            self.max_digits = self.field_info.max_digits
            self.decimal_places = self.field_info.decimal_places

        # Validation on precision and scale per Avro schema
        if self.max_digits <= 0:
            raise ValueError("`max_digits` must be a positive integer greater than zero")

        if self.decimal_places < 0 or self.max_digits < self.decimal_places:
            raise ValueError("`decimal_places` must be zero or a positive integer less than or equal to the precision.")

    def get_avro_type(
        self,
    ) -> typing.Union[types.JsonDict, typing.List[typing.Union[str, types.JsonDict]]]:
        avro_type = {
            "type": field_utils.BYTES,
            "logicalType": field_utils.DECIMAL,
            "precision": self.max_digits,
            "scale": self.decimal_places,
        }

        if self.default is None:
            return ["null", avro_type]

        return avro_type

    def default_to_avro(self, default: decimal.Decimal) -> str:
        return serialization.decimal_to_str(default, self.max_digits, self.decimal_places)

    def fake(self) -> decimal.Decimal:
        return fake.pydecimal(
            right_digits=self.decimal_places,
            left_digits=self.max_digits - self.decimal_places,
        )


@dataclasses.dataclass
class RecordField(Field):
    def get_avro_type(self) -> typing.Union[str, typing.List, typing.Dict]:
        meta = getattr(self.type, "Meta", type)
        metadata = utils.SchemaMetadata.create(meta)

        alias = self.parent._metadata.get_alias_nested_items(self.name) or metadata.get_alias_nested_items(self.name)  # type: ignore  # noqa E501

        # The priority for the schema name
        # 1. Check if exists an alias_nested_items in parent llass or Meta class of own model
        # 2. Check if the schema_name is present in the Meta class of own model
        # 3. Use the default class Name (self.type.__name__)
        name = alias or metadata.schema_name or self.type.__name__

        if not self.exist_type() or alias is not None:
            user_defined_type = utils.UserDefinedType(name=name, type=self.type)
            self.parent._user_defined_types.add(user_defined_type)

            record_type = self.type.avro_schema_to_python(parent=self.parent)
            record_type["name"] = name
        else:
            if metadata.namespace is None:
                record_type = name
            else:
                record_type = f"{metadata.namespace}.{name}"

        if self.default is None:
            return [field_utils.NULL, record_type]

        return record_type

    def default_to_avro(self, value: "schema_generator.AvroModel") -> typing.Dict:
        parser = value._parser or value._generate_parser()
        return {
            fieldname: field.default_to_avro(getattr(value, fieldname))
            for fieldname, field in parser.get_fields_map().items()
        }

    def fake(self) -> typing.Any:
        return self.type.fake()


from .mapper import (
    ALL_TYPES_FIELD_CLASSES,
    CONTAINER_FIELDS_CLASSES,
    IMMUTABLE_FIELDS_CLASSES,
    LOGICAL_TYPES_FIELDS_CLASSES,
    SPECIAL_ANNOTATED_TYPES,
)

LOGICAL_CLASSES = LOGICAL_TYPES_FIELDS_CLASSES.keys()
PYDANTIC_CUSTOM_CLASS_METHOD_NAMES = {
    "__get_validators__",
    "__get_pydantic_core_schema__",
}


def field_factory(
    name: str,
    native_type: typing.Any,
    parent: typing.Any = None,
    *,
    default: typing.Any = dataclasses.MISSING,
    default_factory: typing.Any = dataclasses.MISSING,
    metadata: typing.Optional[typing.Dict[str, typing.Any]] = None,
    model_metadata: typing.Optional[utils.SchemaMetadata] = None,
) -> Field:
    from dataclasses_avroschema import AvroModel

    if model_metadata is None:
        model_metadata = utils.SchemaMetadata()

    if metadata is None:
        metadata = {}

    field_info = None
    if native_type is None:
        native_type = type(None)

    if utils.is_annotated(native_type):
        a_type, *extra_args = get_args(native_type)
        field_info = next((arg for arg in extra_args if isinstance(arg, types.FieldInfo)), None)

        if field_info is not None:
            # it means that it is a custom type defined by us `Int32`, `Float32`,`TimeMicro`, `DateTimeMicro`
            # confixed or condecimal
            native_type = utils.rebuild_annotation(a_type, field_info)

        if native_type not in ALL_TYPES_FIELD_CLASSES:
            # type Annotated with the end user
            native_type = a_type

    if native_type in IMMUTABLE_FIELDS_CLASSES:
        klass = IMMUTABLE_FIELDS_CLASSES[native_type]
        return klass(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )

    # special case for some dynamic pydantic types (especially constraint types)
    # when a type cannot be imported and needs to be referenced by qualified string
    # see pydantic conint() implementation for more information
    elif inspect.isclass(native_type) and f"{native_type.__name__}" in IMMUTABLE_FIELDS_CLASSES:
        klass = IMMUTABLE_FIELDS_CLASSES[f"{native_type.__name__}"]
        return klass(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )

    elif utils.is_self_referenced(native_type, parent):
        return SelfReferenceField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif native_type in (types.Fixed, decimal.Decimal):
        klass = SPECIAL_ANNOTATED_TYPES[native_type]  # type: ignore
        return klass(  # type: ignore
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
            field_info=field_info,
        )
    elif native_type in LOGICAL_TYPES_FIELDS_CLASSES:
        klass = LOGICAL_TYPES_FIELDS_CLASSES[native_type]  # type: ignore

        return klass(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif isinstance(native_type, GenericAlias):  # type: ignore
        origin = get_origin(native_type)

        if origin in CONTAINER_FIELDS_CLASSES:
            container_klass = CONTAINER_FIELDS_CLASSES[origin]
        elif origin is typing.Literal:
            container_klass = LiteralField
        else:
            raise ValueError(
                f"Invalid Type {native_type} for field {name}. "
                "Accepted types are list, tuple, dict, typing.Union, or typing.Literal"
            )

        # check here
        return container_klass(  # type: ignore
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif inspect.isclass(native_type) and issubclass(native_type, enum.Enum):
        return EnumField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif types.UnionType is not None and isinstance(native_type, types.UnionType):
        # we need to check whether types.UnionType because it works only in
        # python 3.9 or importing __future__ in previous python versions
        # cases when a container is used, for example `typing.List[int] | str` in python is
        # translated to typing.Union[typing.List[int], str] so it won't reach this point
        return UnionField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    elif inspect.isclass(native_type) and issubclass(native_type, AvroModel):
        return RecordField(
            name=name,
            type=native_type,
            default=default,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
            parent=parent,
        )
    # See if this is a pydantic "Custom Class"
    elif (
        inspect.isclass(native_type)
        and not is_pydantic_model(native_type)  # type: ignore[arg-type]
        and any(method_name in dir(native_type) for method_name in PYDANTIC_CUSTOM_CLASS_METHOD_NAMES)
    ):
        if getattr(parent, "__config__", None):
            try:
                # Build a field for the encoded type since that's what will be serialized
                encoded_type = parent.__config__.json_encoders[native_type]
            except KeyError:
                raise ValueError(
                    f"Type {native_type} for field {name} must be "
                    "listed in the pydantic 'json_encoders' config for {parent}"
                    " (or for one of the classes in its inheritance tree since "
                    "pydantic configs are inherited)"
                )
        else:
            encoded_type = parent.model_config["json_encoders"][native_type]

        # default_factory is not schema-friendly for Custom Classes since it could be returning
        # dynamically constructed values that should not be treated as defaults. For example,
        # native_type could be a timestamp field with a default factory that returns the current
        # time, which would result in "generate_schema" producing a schema with a different default
        # everytime that it's called from AvroModel.
        default_factory = dataclasses.MISSING

        # Encode the default value if it's an instance of native_type
        default_value = encoded_type(default) if type(default) is native_type else default

        # Build a field for the encoded type
        return field_factory(
            name,
            native_type=encoded_type,
            parent=parent,
            default=default_value,
            default_factory=default_factory,
            metadata=metadata,
            model_metadata=model_metadata,
        )
    else:
        msg = (
            f"Type {native_type} for field {name} is unknown. Please check the valid types at "
            "https://marcosschroh.github.io/dataclasses-avroschema/fields_specification/#avro-field-and-python-types-summary"  # noqa: E501
        )

        raise ValueError(msg)


AvroField = field_factory
